module Char
  ( Character(..)
  , char
  , Sword(..)
  , Namespace(..)
  , Effect(..)
  , EffectContext(..)
  , RichEffect(..)
  , PrimEffect(..)
  , ActionChain(..)
  , SwordRef(..)
  , CharacterState(..)
  , CharacterInfo(..)
  , StatusBad(..)
  , DefenceStatus(..)
  , GuardStatus(..)
  , ReviveStatus(..)
  , StatusOther(..)
  , SwordState(..)
  , LogCommand(..)
  , TeamRef(..)
  , GameState(..)
  , GameResult(..)

  -- re-exports
  , CharRef(..)
  ) where

import {-# SOURCE #-} Game
import Bag
import CharRef(CharRef(..), CharMap)

import Data.List
import Data.Maybe
import qualified Data.Map as M
import Data.Monoid

data Character = Ch
  { chName :: String
  , chHp :: Integer
  , chAtk :: Integer
  , chDef :: Integer
  , chSpd :: Integer
  , chSwords :: [Sword]
  }
  deriving (Eq, Ord)

char :: Integer -> Integer -> Integer -> Integer -> [Sword] -> String -> Character
char hp atk def spd swords name = Ch{ chName = name, chHp = hp, chAtk = atk, chDef = def, chSpd = spd, chSwords = swords }

instance Show Character where
  show Ch{ chName = name, chHp = hp, chAtk = atk, chDef = def, chSpd = spd, chSwords = swords } =
    widen 2 (show hp) ++ "/" ++ show atk ++ "/" ++ show def ++ "/" ++ show spd ++ "/" ++ concatMap swShortName swords ++ "/" ++ name
    where
      widen n s = replicate (n - length s) ' ' ++ s

data Sword = Sw
  { swName :: String
  , swNamespace :: Namespace
  , swShortName :: String
  , swEffect :: Effect
  , swTurn :: Int
  , swLimited :: Bool
  , swPeriod :: Maybe Int
  , swMP :: Maybe Integer
  , swReactsToDanger :: Bool
  , swReactsToIllness :: Bool
  }

data Namespace = M100 | M2011a | Private
  deriving (Show, Eq, Ord)

instance Show Sword where
  show Sw{ swName = name } = name

instance Eq Sword where
  x == y = swShortName x == swShortName y && swNamespace x == swNamespace y

instance Ord Sword where
  compare x y = 
    compare (swNamespace x) (swNamespace y) `mappend`
    compare (swShortName x) (swShortName y)

data Effect 
  = EAtk [RichEffect]
  | EAux [RichEffect]
  | ECtl Bool (GameState -> EffectContext -> SwordState -> (Maybe SwordState, [String]))
  | EAdd (GameState -> EffectContext -> Maybe SwordState)
  | EBld (GameState -> EffectContext -> Character)
  | EEnv (GameState -> EffectContext -> Game ())

data EffectContext = EC
  { ecThisSword :: SwordRef
  , ecNextSword :: Maybe SwordRef
  , ecOwner :: CharRef
  }

data RichEffect = RE
  { rePrim :: PrimEffect
  , reRepeat :: Int
  }

data PrimEffect
  = TargetEnemy
    { teDefendable :: Bool
    , teAll :: Bool
    , teGen :: CharRef -> EffectContext -> ActionChain
    }
  | TargetAlly
    { taAll :: Bool
    , taGen :: CharRef -> EffectContext -> ActionChain
    }
  | TargetNone
    { tnGen :: EffectContext -> ActionChain
    }

newtype ActionChain = AC { unAC :: GameState -> Game ([CharRef], Maybe ActionChain) }
  -- returns a list of the damaged characters, and 'AND' clause if present.

newtype SwordRef = SR Int
  deriving (Eq, Ord, Show)

data CharacterState = CS
  { csDead :: !Bool
  , csMaxHp :: !Integer
  , csHp :: !Integer
  , csAtk :: !Integer
  , csDef :: !Integer
  , csSpd :: !Integer
  , csSwords :: ![SwordRef]
  , csAp :: !Integer
  , csReservedAp :: !Integer
  , csMp :: !Integer
  , csStatusBad :: !StatusBad
  , csStatusDefence :: !(Bag DefenceStatus)
  , csStatusGuard :: !(Bag GuardStatus)
  , csStatusRevive :: !(Bag ReviveStatus)
  , csStatusOther :: !StatusOther
  , csSwordState :: !(M.Map SwordRef SwordState)
  , csSwordRefSupply :: !Int
  , csInfo :: !CharacterInfo
  }
  deriving (Eq)

data CharacterInfo = CI
  { ciDispName :: !String
  , ciSide :: !TeamRef
  , ciTurnBuilt :: !(Maybe Int)
  , ciOriginal :: !Character
  }
  deriving (Eq)

data StatusBad = SB
  { sbStun :: !Int
  , sbSleep :: !Int
  , sbBadPoison :: !Int
  , sbPoison :: !Int
  , sbAbortion :: !Int
  }
  deriving (Eq)

data DefenceStatus = Reflect | Detect
  deriving (Eq, Ord)

instance Show DefenceStatus where
  show ds = case ds of
    Reflect -> "反射"
    Detect -> "見切り"

data GuardStatus = Barrier | Cusion | Flip | Shield | Guard
  deriving (Eq, Ord)

instance Show GuardStatus where
  show gs = case gs of
    Barrier -> "バリア"
    Cusion -> "クッション"
    Flip -> "フリップ"
    Shield -> "シールド"
    Guard -> "ガード"

data ReviveStatus = Soul | Life
  deriving (Eq, Ord)

instance Show ReviveStatus where
  show rs = case rs of
    Soul -> "ソウル"
    Life -> "ライフ"

data StatusOther = SO
  { soPush :: !Int
  , soHeavensDoor :: !Bool
  , soQuick :: !Int
  , soSlow :: !Int
  , soDrive :: !Int
  , soProvoke :: !Bool
  }
  deriving (Eq)

instance Show CharacterState where
  show CS{ csInfo = info, csMaxHp = maxhp, csHp = hp, csAtk = atk, csDef = def,
          csSpd = spd, csSwords = swords, csAp = ap, csReservedAp = rap, csStatusBad = bad,
          csStatusGuard = guard, csStatusDefence = defence, csStatusRevive = revive,
          csStatusOther = other, csSwordState = sstate, csMp = mp } =
    unwords $ [ciDispName info, status] ++ swordline ++ points ++ effects
    where
      status = show hp ++ "(" ++ show maxhp ++ ")/" ++ show atk ++ "/" ++ show def ++ "/" ++ show spd
      swordline = map (show . (sstate M.!))  swords
      effects = concatMap manyAttr $ bads ++ pr guard ++ pr defence ++ pr revive ++ others
      pr b = map (\(x, y) -> (show x, fromIntegral y)) $ assocs b
      bads = case bad of
        SB{ sbPoison = poison, sbSleep = sleep, sbStun = stun, sbBadPoison = bp, sbAbortion = ab } ->
          [("スタン", stun), ("スリープ", sleep), ("猛毒", bp), ("毒", poison), ("流産", ab)]
      others = case other of
        SO{ soProvoke = provoke, soHeavensDoor = heaven, soPush = push, soQuick = quick, soDrive = drive, soSlow = slow } ->
          [("プッシュ", push), ("ヘヴンズドア", b2c heaven), ("クイック", quick), ("スロー", slow), ("ドライブ", drive), ("挑発", b2c provoke)]

      b2c b = if b then 1 else 0
      points = mapMaybe fmtPoint [("MP", mp), ("AP", ap), ("予備AP", rap)]
      fmtPoint (name, 0) = Nothing
      fmtPoint (name, val) = Just $ name ++ ":" ++ show val

      manyAttr :: (String, Int) -> [String]
      manyAttr (sym, 0) = []
      manyAttr (sym, 1) = ["<" ++ sym ++ ">"]
      manyAttr (sym, 2) = ["<" ++ sym ++ ">", "<" ++ sym ++ ">"]
      manyAttr (sym, n) = ["<" ++ sym ++ ">*" ++ show n]

data SwordState = SS
  { ssSword :: Sword
  , ssUsed :: Bool
  , ssStun :: Int
  , ssPoison :: Int
  , ssAll :: Bool
  , ssEarly :: Int
  , ssLimitAdded :: Bool
  , ssForcedReactsToDanger :: Bool
  }
  deriving (Eq)

instance Show SwordState where
  show SS{ ssSword = s, ssUsed = u, ssStun = stun, ssPoison = poison, ssAll = al, ssEarly = e, ssLimitAdded = lim }
    = p $ swName s ++ rep (-e) "<遅>" ++ rep e "<早>" ++ rep stun "<振>" ++ rep poison "<与毒>" ++ repb al "<全>" ++ repb lim "<リミ>"
    where
      p x
        | u = "(" ++ x ++ ")"
        | otherwise = x
      rep n k = concat $ replicate n k
      repb b = rep $ if b then 1 else 0

data LogCommand = PutLine String | PutNoIndent String | ChangeIndent Int

data TeamRef = Team0 | Team1
  deriving (Show, Eq, Ord)

data GameState = GS
  { gsTurnCount :: !Int
  , gsUniqueSupply0 :: !Int
  , gsUniqueSupply1 :: !Int
  , gsChars :: !(CharMap CharacterState)
  , gsTeam0 :: ![CharRef]
  , gsTeam1 :: ![CharRef]
  , gsSomeoneMayBeDead :: !Bool -- for optimization. true when it is possible that one or more present chars are dead
  , gsSwordTransfer :: !(M.Map (CharRef, SwordRef) (CharRef, SwordRef))
  }
  deriving (Show, Eq)

newtype GameResult = GR
  { grWinner :: Maybe TeamRef
  }
  deriving (Show, Eq)
