module Parse
  ( parseMaaguwerite
  , parseStst
  , parseCharacter
  , parseMaaguweriteG
  , parseStstG
  , parseCharacterG
  , looksLikeCharacterSpec
  ) where

import Char
import Swords

import Data.Maybe
import Data.Char
import Data.List
import Control.Monad
import Text.ParserCombinators.ReadP as ReadP

looksLikeCharacterSpec :: String -> Bool
looksLikeCharacterSpec str = not $ null $ readP_to_S p str
  where
    p = many get >> natP >> skipSpaces >> slashP >> skipSpaces >> natP >> skipSpaces >> slashP

runReadS :: ReadS a -> String -> Maybe a
runReadS rs input = listToMaybe $ map fst $ rs input

runReadP :: ReadP a -> String -> Maybe a
runReadP rp input = runReadS (readP_to_S rp') input
  where
    rp' = do
      r <- rp
      eofP
      return r

parseCharacter, parseStst, parseMaaguwerite :: String -> Maybe Character
parseCharacter = parseCharacterG M100
parseStst = parseStstG M100
parseMaaguwerite = parseMaaguweriteG M100

parseCharacterG, parseStstG, parseMaaguweriteG :: Namespace -> String -> Maybe Character
parseCharacterG pref = runReadP (maaguweriteGP pref <++ ststGP pref)
parseStstG pref = runReadP (ststGP pref)
parseMaaguweriteG pref = runReadP (maaguweriteGP pref)

type Par = ReadP

maaguweriteGP :: Namespace -> Par Character
maaguweriteGP pref = do
  hp <- natP
  [atk, def, spd] <- count 3 $ slashP >> natP
  sw <- slashP >> many (qualifiedP pref swordShortP)
  name <- option "" $ slashP >> many get
  return $ Char.char hp atk def spd sw name

slashP :: Par ()
slashP = ReadP.char '/' >> return ()

ststGP :: Namespace -> Par Character
ststGP pref = do
  name <- option "" $ do
    s <- option "" $ do
      ini <- many get
      las <- satisfy $ not . isSpace
      return $ ini ++ [las]
    ss $ anyOfP "：:"
    return s
  hp <- ss natP
  [atk, def, spd] <- count 3 $ ss slashP >> ss natP
  ss slashP
  sws <- ss (qualifiedP pref swordLongP) `sepBy` ss (anyOfP "、,，")
  return $ Char.char hp atk def spd sws name
  where
    ss = (skipSpaces>>)

anyOfP :: [Char] -> Par ()
anyOfP xs = do
  choice $ map ReadP.char xs
  return ()

qualifiedP :: Namespace -> (Namespace -> Par a) -> Par a
qualifiedP ns action = do
  ns' <- option ns $ do
    ReadP.char '<'
    ns' <- namespaceP
    ReadP.char '>'
    return ns'
  action ns'

namespaceP :: Par Namespace
namespaceP = mapP [("100", M100), ("1a", M2011a), ("kotha.net", Private)]

swordShortP :: Namespace -> Par Sword
swordShortP ns = swordP ns $ map shrt allSwords ++ [("デ", death), ("コ", ko)] ++ map (\n -> ("遅" ++ show n, delay n)) [0..20]
  where
    shrt s = (swShortName s, s)

swordLongP :: Namespace -> Par Sword
swordLongP ns = swordP ns $ map lng allSwords
  where
    lng s = (swName s, s)

type SwordDict = [(String, Sword)]

swordP :: Namespace -> SwordDict -> Par Sword
swordP ns dict = mapP a <++ mapP b
  where (a, b) = partition ((==ns) . swNamespace . snd) dict

mapP :: [(String, a)] -> Par a
mapP m = choice $ map constantP m

constantP :: (String, a) -> Par a
constantP (name, sw) = string name >> return sw

natP :: Par Integer
natP = fmap read $ many1 $ satisfy isDigit

eofP :: Par ()
eofP = look >>= guard . null
