module Util where

import Control.Exception
import Control.Monad
import Data.List
import Data.Maybe
import System.Environment
import System.IO.Error(isDoesNotExistError)

pair :: (a -> b) -> a -> (a, b)
pair f x = (x, f x)

whenJust :: (Monad m) => Maybe a -> (a -> m ()) -> m ()
whenJust m f = maybe (return ()) f m

data Schwartz k v = Schwartz !k v

schwartz :: (v -> k) -> v -> Schwartz k v
schwartz key val = Schwartz (key val) val

unSchwartz :: Schwartz k v -> v
unSchwartz (Schwartz _ val) = val

schwartzKey :: Schwartz k v -> k
schwartzKey (Schwartz key _) = key

instance (Eq k) => Eq (Schwartz k v) where
  Schwartz k0 _ == Schwartz k1 _ = k0 == k1

instance (Ord k) => Ord (Schwartz k v) where
  compare (Schwartz k0 _) (Schwartz k1 _) = compare k0 k1

sortOn :: (Ord k) => (a -> k) -> [a] -> [a]
sortOn f = map unSchwartz . sort . map (schwartz f)

minsOn, maxesOn :: (Ord k) => (a -> k) -> [a] -> [a]
minsOn = minsOrMaxes LT
maxesOn = minsOrMaxes GT

minsOrMaxes :: (Ord k) => Ordering -> (a -> k) -> [a] -> [a]
minsOrMaxes _ _ [] = []
minsOrMaxes cond f (x:xs) = loop [x] (f x) xs
  where
    loop cur now [] = reverse cur
    loop cur now (y:ys)
      | od == cond = loop [y] val ys
      | od == EQ = loop (y:cur) now ys
      | otherwise = loop cur now ys
      where
        od = compare val now
        val = f y

sortGroupOn, revSortGroupOn :: (Ord k) => (a -> k) -> [a] -> [[a]]
sortGroupOn f = map (map unSchwartz) . group . sort . map (schwartz f)
revSortGroupOn f = map (map unSchwartz) . group . sortBy (flip compare) . map (schwartz f)

head1 [] = []
head1 (x:xs) = x

zipMaybe :: [a] -> [b] -> [(Maybe a, Maybe b)]
zipMaybe xs ys = takeWhile valid $ zip (extend xs) (extend ys)
  where
    extend list = map Just list ++ repeat Nothing
    valid (x, y) = isJust x || isJust y

sortNub :: (Ord a) => [a] -> [a]
sortNub = uniq . sort

uniq :: (Eq a) => [a] -> [a]
uniq [] = []
uniq (x:xs) = x : map fst (filter (uncurry (/=)) $ zip xs (x:xs))

getEnvMaybe :: String -> IO (Maybe String)
getEnvMaybe key = handleJust (guard . isDoesNotExistError) (const $ return Nothing) (fmap Just $ getEnv key)

removeByIndices :: [Int] -> [a] -> [a]
removeByIndices = loop 0
  where
    loop idx [] xs = xs
    loop idx _ [] = []
    loop idx (k:ks) (x:xs) = case compare idx k of
      LT -> x : loop (idx+1) (k:ks) xs
      GT -> loop idx ks (x:xs)
      EQ -> loop (idx+1) ks xs

