
# include <memory>
# include <algorithm>
# include <sstream>
# include "boost/preprocessor/repetition/repeat.hpp"
# include "boost/preprocessor/tuple/rem.hpp"
# include "boost/preprocessor/seq/transform.hpp"
# include "boost/preprocessor/cat.hpp"
# include "boost/lambda/lambda.hpp"
# include "boost/lambda/bind.hpp"
# include "boost/lambda/construct.hpp"

# include "internal.h"

using namespace m100sim;
namespace bll = boost::lambda;

namespace
{

struct sinfo_builder
{
  sinfo_builder(const effect &eff, int prio)
    : eff(eff), prio(prio) {}
  sinfo build(const char *name, const char *shortname, int turn = 1, bool limited = false,
      optional<int> mp = boost::none, optional<int> period = boost::none,
      bool danger = false, bool illness = false) const
  {
    return sinfo(name, shortname, eff, prio, turn, limited, mp, period, danger, illness);
  }
private:
  effect eff;
  int prio;
};

effect single_effect(const rich_effect &re)
{
  return effect(1, re);
}

effect double_effect(const rich_effect &re0, const rich_effect &re1)
{
  effect ret;
  ret.push_back(re0);
  ret.push_back(re1);
  return ret;
}

template <typename Fun>
struct simple_prim_effect : prim_effect_base
{
  simple_prim_effect(const Fun &fun, int stage) : fun(fun), count(stage) {}
  bool perform(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
  {
    if(count-- > 0)
      return true;
    fun(gs, ctx, target, lg);
    return false;
  }
private:
  const Fun fun;
  int count;
};

template <typename Fun>
boost::function<prim_effect_base *()> simple_prim(const Fun &fun, int stage = 1)
{
  return bll::bind(bll::new_ptr<simple_prim_effect<Fun> >(), fun, stage);
}

template <typename Fun>
rich_effect simple_re(const Fun &fun)
{
  return rich_effect(target_none, simple_prim(fun));
}

template <typename T, typename Getter, typename Modifier>
struct general_gain_functor
{
  general_gain_functor(const char *desc, const Getter &get, const Modifier &mod, bool reregister = false, bool revive = false)
    :desc(desc), get(get), mod(mod), rereg(reregister), rev(revive) {}
  void operator()(gstate &gs, const effect_context &ctx, cstate *target, logger &lg) const
  {
    T old_v;
    if(rereg) unregister_from_action_order(gs, target);
    if(!lg.null()) old_v = get(target);
    mod(target);
    if(FILE *out = lg.start_direct())
    {
      fprintf(out, "%sは%sを得た。", target->info.name.c_str(), desc);
      print_int(out, old_v);
      fputs(" -> ", out);
      print_int(out, get(target));
      lg.end_direct();
    }
    if(rereg) register_to_action_order(gs, target);
    if(rev && target->dead && invoke_revival(gs, target, lg))
      revive(gs, target, lg);
  }
private:
  const char *desc;
  T cstate::*pmem;
  Getter get;
  Modifier mod;
  bool rereg;
  bool rev;
};

template <typename T, typename Getter, typename Modifier>
general_gain_functor<T, Getter, Modifier> make_general_gain_functor(const char *desc, const Getter &get, const Modifier &mod)
{
  return general_gain_functor<T, Getter, Modifier>(desc, get, mod);
}

template <typename T, typename Getter, typename Modifier>
boost::function<prim_effect_base *()> general_gain(const char *desc, const Getter &get, const Modifier &mod, bool rereg, bool rev = false)
{
  return simple_prim(general_gain_functor<T, Getter, Modifier>(desc, get, mod, rereg, rev));
}

template <typename T>
boost::function<prim_effect_base *()> gain1(const char *desc, T cstate::*pmem, int incr, bool rereg = false)
{
  return general_gain<T>(desc, bll::bind(pmem, bll::_1), bll::bind(pmem, bll::_1) += bll::constant(incr), rereg);
}

void put_bad(gstate &gs, const effect_context &ctx, const char *desc, int status_bad::*pmem, cstate *target, logger &lg)
{
  make_general_gain_functor<int>(desc, 
      bll::bind(pmem, bll::bind(&cstate::bad, bll::_1)),
      ++bll::bind(pmem, bll::bind(&cstate::bad, bll::_1)))
    (gs, ctx, target, lg);
}

int_t handle_guard(cstate *tgt, intparam dmg, logger &lg)
{
  multiset<guard_status> &gs = tgt->guard;
  multiset<guard_status>::iterator fst = gs.begin();
  if(fst == gs.end()) return dmg;
  guard_status s = *fst;
  gs.erase(fst);
  int_t newdmg;
  if(s == flip && tgt->def > 0)
  {
    int_t newdef = tgt->def - 1;
    newdmg = dmg - 3;
    lg.put(boost::format("%1%のフリップが発動した。防: %2% -> %3%、ダメージ: %4% -> %5%")
        % tgt->info.name % tgt->def % newdef % dmg % newdmg);
    tgt->def = newdef;
    return newdmg;
  }
  switch(s)
  {
  case barrier:
    newdmg = 0;
    break;
  case cusion:
    newdmg = 1;
    break;
  case flip:
    newdmg = dmg;
    break;
  case shield:
    newdmg = div_floor(dmg, 2);
    break;
  case guard:
    newdmg = div_floor(dmg + 1, 2);
    break;
  }
  lg.put(boost::format("%1%の%2%が発動した。ダメージ: %3% -> %4%") % tgt->info.name % s % dmg % newdmg);
  return newdmg;
}

int_t handle_drive(cstate *self, intparam dmg, logger &lg)
{
  if(!self->other.drive) return dmg;
  self->other.drive--;
  if(self->atk != 0)
  {
    int_t newdmg = dmg + 3;
    int_t newatk = self->atk - 1;
    lg.put(boost::format("%1%のドライブが発動した。攻: %2% -> %3%、ダメージ: %4% -> %5%") % self->info.name % self->atk % newatk % dmg % newdmg);
    self->atk = newatk;
    return newdmg;
  }
  else
  {
    lg.put(boost::format("%1%のドライブが発動した。攻: %2% -> %2%、ダメージ: %3% -> %3%") % self->info.name % self->atk % dmg);
    return dmg;
  }
}

int_t declare_damage(gstate &gs, cstate *self, cstate *tgt, intparam dmg, bool penetrate, logger &lg)
{
  int_t dz = handle_drive(self, dmg, lg);
  return penetrate ? dz : handle_guard(tgt, dz, lg);
}

void mp_damage(cstate *tgt, intparam amount, logger &lg)
{
  if(amount <= 0) return;
  int_t newmp = tgt->mp - amount;
  if(FILE *out = lg.start_direct())
  {
    fprintf(out, "%sは", tgt->info.name.c_str());
    print_int(out, amount);
    fputs("点のMPダメージを受けた。MP: ", out);
    print_int(out, tgt->mp);
    fputs(" -> ", out);
    print_int(out, newmp);
    lg.end_direct();
  }
  tgt->mp = newmp;
}

void attack_damage(gstate &gs, cstate *tgt, intparam amount, logger &lg, bool mp = false)
{
  if(amount <= 0)
    lg.putf("%sはダメージを受けなかった", tgt->info.name.c_str());
  else if(mp)
    mp_damage(tgt, amount, lg);
  else
    damage(gs, "攻撃", tgt, amount, lg);
}

struct recoverer
{
  recoverer(cstate *tgt, intparam amt)
    : tgt(tgt), amt(amt) {}
  void operator()(gstate &gs, logger &lg) const
  {
    int_t new_hp = max<int_t>(tgt->hp, min<int_t>(tgt->max_hp, tgt->hp + amt));
    if(new_hp <= tgt->hp) return;
    lg.put(boost::format("%1%は%2%点回復した。HP: %3% -> %4%")
        % tgt->info.name % (new_hp-tgt->hp) % tgt->hp % new_hp);
    set_hp(gs, tgt, new_hp, lg);
  }
private:
  cstate *tgt;
  int_t amt;
};

struct sword_stealer
{
  explicit sword_stealer(cstate *self, cstate *tgt)
    :self(self), tgt(tgt), swd(tgt->swords.begin()) {}
  void operator()(gstate &gs, logger &lg) const
  {
    if(swd == tgt->swords.end()) return;
    if(FILE *out = lg.start_direct())
    {
      fprintf(out, "%sは%sの", self->info.name.c_str(), tgt->info.name.c_str());
      print_sstate(out, *swd);
      fputs("を盗んだ", out);
      lg.end_direct();
    }
    self->swords.splice(self->swords.end(), tgt->swords, swd);
  }
private:
  cstate *self;
  cstate *tgt;
  sref swd;
};

typedef int_t (*ptr_damage_fun)(int, const cstate *, const cstate *);
typedef void (*ptr_effect_fun)(gstate &, const effect_context &, cstate *, logger &);
typedef boost::function<void (gstate &, logger &)> action;
typedef action (*ptr_next_fun)(gstate &, intparam dmg, const effect_context &, cstate *, logger &);

void no_action(gstate &, logger &) {}

struct attack_options
{
  explicit attack_options(ptr_damage_fun dmg)
    : dmg(dmg), eff(0), next(0), self(false), count(1), scatter(false), penetrate(false), mp(false) {}
  ptr_damage_fun dmg;
  ptr_effect_fun eff;
  ptr_next_fun next;
  bool self;
  int count;
  bool scatter;
  bool penetrate;
  bool mp;
};

struct attack_prim_effect : prim_effect_base
{
  explicit attack_prim_effect(const attack_options &opts) : opts(opts), stage(0) {}
  bool perform(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
  {
    switch(stage++)
    {
    case 0:
      dmg_val = opts.dmg(gs.turn, ctx.owner, target);
      return true;
    case 1:
      dmg_val = declare_damage(gs, ctx.owner, target, dmg_val, opts.penetrate, lg);
      return true;
    case 2:
      attack_damage(gs, target, dmg_val, lg, opts.mp);
      if(dmg_val > 0)
      {
        if(opts.eff)
          opts.eff(gs, ctx, target, lg);
        for(int i = 0; i < ctx.ench.stun; i++)
          put_bad(gs, ctx, "スタン", &status_bad::stun, target, lg);
        for(int i = 0; i < ctx.ench.poison; i++)
          put_bad(gs, ctx, "毒", &status_bad::poison, target, lg);
      }
      return opts.next;
    case 3:
      next = opts.next(gs, dmg_val, ctx, target, lg);
      return true;
    case 4:
      next(gs, lg);
      return false;
    }
  }
private:
  const attack_options &opts;
  int stage;
  int_t dmg_val;
  boost::function<void (gstate &, logger &)> next;
};

void add_sword(gstate &gs, cstate *self, const sstate &to_add, logger &lg)
{
  self->swords.push_back(to_add);
  self->swords.back().used = false;
  if(FILE *out = lg.start_direct())
  {
    fprintf(out, "%sは", self->info.name.c_str());
    print_sstate(out, self->swords.back());
    fputs("を得た", out);
    lg.end_direct();
  }
}

typedef boost::optional<sstate> (*ptr_sword_chooser)(gstate &, const effect_context &ctx, logger &);

struct add_prim_effect : prim_effect_base
{
  add_prim_effect(ptr_sword_chooser choose) : choose(choose), stage(0) {}
  bool perform(gstate &gs, const effect_context &ctx, cstate *, logger &lg)
  {
    switch(stage++)
    {
    case 0:
      to_add = choose(gs, ctx, lg);
      return to_add;
    case 1:
      add_sword(gs, ctx.owner, *to_add, lg);
      return false;
    }
  }
private:
  ptr_sword_chooser choose;
  int stage;
  optional<sstate> to_add;
};


effect add_effect(ptr_sword_chooser choose)
{
  return single_effect(rich_effect(target_none, bll::bind(bll::new_ptr<add_prim_effect>(), choose)));
}

typedef const m100sim_character *(*ptr_build_chooser)(gstate &gs, const effect_context &ctx);

void build(gstate &gs, cstate *self, const m100sim_character &to_build, logger &lg)
{
  const char *fail_reason = NULL;
  if(self->bad.abortion > 0)
  {
    self->bad.abortion--;
    fail_reason = "流産が発動したため";
  }
  else if(self->info.turn_built == gs.turn)
    fail_reason = "生まれたてのため";
  if(fail_reason)
  {
    lg.putf("%s、%sはビルドに失敗した", fail_reason, self->info.name.c_str());
    return;
  }
  side_id side = self->info.side;
  tstate &team = gs.teams[side];
  ostringstream namestr;
  namestr << (to_build.name ? to_build.name : "") << '/' << 2*team.unique_source + side;
  string name = namestr.str();
  gs.teams[side].chs.push_front(cstate(cinfo(to_build, name, side, team.unique_source++, gs.turn)));
  cstate &cs = gs.teams[side].chs.front();
  cs.ap = cs.spd;
  register_to_action_order(gs, &cs);
  lg.putf("%sがビルドされた", name.c_str());
  if(FILE *out = lg.start_direct())
  {
    print_cstate(out, cs);
    lg.end_direct();
  }
  if(cs.hp <= 0)
    kill(gs, &cs, lg);
}

struct builder
{
  builder(ptr_build_chooser choose) : choose(choose) {}
  void operator()(gstate &gs, const effect_context &ctx, cstate *, logger &lg) const
  {
    build(gs, ctx.owner, *choose(gs, ctx), lg);
  }
private:
  ptr_build_chooser choose;
};

effect build_effect(ptr_build_chooser choose)
{
  return single_effect(simple_re(builder(choose)));
}

rich_effect simple_attack(const attack_options &opts)
{
  return rich_effect(
      opts.self ? target_none : target_enemy,
      bll::bind(bll::new_ptr<attack_prim_effect>(), bll::constant_ref(opts)),
      !opts.self && !opts.penetrate,
      opts.count,
      opts.scatter);
}

template <typename T>
struct cstate_component
{
  static T cstate::*const value;
};

template <> status_bad cstate::*const cstate_component<status_bad>::value = &cstate::bad;
template <> status_other cstate::*const cstate_component<status_other>::value = &cstate::other;
template <> multiset<guard_status> cstate::*const cstate_component<multiset<guard_status> >::value = &cstate::guard;
template <> multiset<defence_status> cstate::*const cstate_component<multiset<defence_status> >::value = &cstate::defence;
template <> multiset<revival_status> cstate::*const cstate_component<multiset<revival_status> >::value = &cstate::revival;

template <typename T>
boost::function<prim_effect_base *()> gain_o(T t, bool rereg = false)
{
  typedef multiset<T> ms;
  ms cstate::*pmem = cstate_component<ms>::value;
  return general_gain<size_t>(name(t),
      bll::bind(&multiset<T>::count, bll::bind(pmem, bll::_1), t),
      bll::bind(static_cast<typename multiset<T>::iterator (multiset<T>::*)(const T&)>(&multiset<T>::insert), &bll::bind(pmem, bll::_1), t),
      rereg);
}

boost::function<prim_effect_base *()> gain_rev(revival_status rev)
{
  typedef multiset<revival_status> ms;
  ms cstate::*pmem = &cstate::revival;
  return general_gain<size_t>(name(rev),
      bll::bind(&ms::count, bll::bind(pmem, bll::_1), rev),
      bll::bind(static_cast<ms::iterator (ms::*)(const revival_status&)>(&ms::insert), &bll::bind(pmem, bll::_1), rev),
      false,
      true);
}

template <typename T>
boost::function<prim_effect_base *()> gain_os(const char *desc, T status_other::*pmem, bool rereg = false)
{
  return general_gain<T>(desc, bll::bind(pmem, bll::bind(&cstate::other, bll::_1)), ++bll::bind(pmem, bll::bind(&cstate::other, bll::_1)), rereg);
}

void recover_full(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
{
  lg.put(boost::format("%1%は全回復した。HP: %2% -> %3%") 
      % target->info.name % target->hp % target->max_hp);
  set_hp(gs, target, target->max_hp, lg);
}

void unuse_self(gstate &gs, const effect_context &ctx, cstate *, logger &lg)
{
  if(ctx.this_sword)
    ctx.this_sword->used = false;
}

void destroy_sword(gstate &gs, cstate *target, logger &lg)
{
  if(target->swords.empty()) return;
  lg.putf("%sの腰骨が折れた", target->info.name.c_str());
  lg.putf("%sの%sは破壊された", target->info.name.c_str(), target->swords.front().info->name);
  remove_sword(gs, target, target->swords.begin());
}

void pressurize(gstate &gs, cstate *target, logger &lg)
{
  sref s = find_if(target->swords.begin(), target->swords.end(), !bll::bind(&sstate::used, bll::_1));
  if(s == target->swords.end()) return;
  lg.putf("%sの%sは使用済みになった", target->info.name.c_str(), s->info->name);
  s->used = true;
}

void reverse_swords(gstate &gs, cstate *target, logger &lg)
{
  lg.putf("%sの剣の順序は逆順になった", target->info.name.c_str());
  target->swords.reverse();
}

struct param_resetter
{
  param_resetter(const char *desc, int_t cstate::*smem, int m100sim_character::*cmem, bool reregister = false)
    : desc(desc), smem(smem), cmem(cmem), rereg(reregister) {}
  void operator()(gstate &gs, const effect_context &ctx, cstate *target, logger &lg) const
  {
    if(rereg) unregister_from_action_order(gs, target);
    lg.put(boost::format("%1%の%2%が初期値に戻った。%3% -> %4%")
        % target->info.name % desc % target->*smem % target->info.original.*cmem);
    target->*smem = target->info.original.*cmem;
    if(rereg) register_to_action_order(gs, target);
  }
private:
  const char *desc;
  int_t cstate::*smem;
  int m100sim_character::*cmem;
  bool rereg;
};

void change_sword_type(sstate *ss, const sinfo *info)
{
  ss->info = info;
  ss->ench = enchantments();
}

void change_sword_type_as(sstate *ss, const sstate &ref)
{
  ss->info = ref.info;
  ss->ench = ref.ench;
}

void do_flashfire(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
{
  lg.putf("%sの剣が全て炸撃剣になった", target->info.name.c_str());
  for(sref it = target->swords.begin(); it != target->swords.end(); ++it)
    change_sword_type(&*it, sinfo_by_ref(m100sim_s_firecracker));
}

void recover_all(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
{
  lg.putf("%sの悪い状態が全て除去された", target->info.name.c_str());
  target->bad = status_bad();
}

void do_prospect(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
{
  lg.put(boost::format("%1%は最大HPとHPとスリープを得た。(%2%,%3%,%4%) -> (%5%,%6%,%7%)")
      % target->info.name % target->max_hp % target->hp % target->bad.sleep
      % (target->max_hp+10) % (target->hp+10) % (target->bad.sleep + 1));
  target->max_hp += 10;
  target->hp += 10;
  target->bad.sleep++;
}

void do_dragonize(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
{
  lg.put(boost::format(
        "%1%の能力が倍になった。最大HP: %2% -> %3%、HP: %4% -> %5%、MP: %6% -> %7%、"
        "攻撃力: %8% -> %9%、防御力: %10% -> %11%、素早さ: %12% -> %13%")
      % target->info.name % target->max_hp % (target->max_hp*2) % target->hp % (2*target->hp) % target->mp % (2*target->mp)
      % target->atk % (2*target->atk) % target->def % (2*target->def) % target->spd % (2*target->spd));
  unregister_from_action_order(gs, target);
  target->max_hp = checked_double(target->max_hp);
  target->hp = checked_double(target->hp);
  target->mp = checked_double(target->mp);
  target->atk = checked_double(target->atk);
  target->def = checked_double(target->def);
  target->spd = checked_double(target->spd);
  register_to_action_order(gs, target);
}

void double_attack(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)
{
  lg.put(boost::format("%1%の攻撃力が倍になった。%2% -> %3%")
      % target->info.name % target->atk % (2*target->atk));
  target->atk = checked_double(target->atk);
}

struct character_builder
{
  character_builder(const char *name, int hp, int atk, int def, int spd)
    : hp(hp), atk(atk), def(def), spd(spd), name(name) {}
  character_builder &operator()(m100sim_sword_ref r) { swords.push_back(r); return *this; }
  character_builder &operator()(boost::none_t) { return *this; }
  void write(m100sim_character *out, vector<m100sim_sword_ref> &vec)
  {
    vec = swords;
    out->name = name;
    out->hp = hp;
    out->atk = atk;
    out->def = def;
    out->spd = spd;
    out->swords = &vec.front();
    out->n_swords = vec.size();
  }
private:
  int hp, atk, def, spd;
  const char *name;
  vector<m100sim_sword_ref> swords;
};

boost::none_t m100sim_s_none; /* hack to avoid empty seq at BLD */
typedef pair<m100sim_character, vector<m100sim_sword_ref> > buildia_storage;

void do_age(gstate &gs, const effect_context &ctx, cstate *, logger &lg)
{
  increment_turn_count(gs, lg);
}

const param_resetter atk_resetter("攻撃力", &cstate::atk, &m100sim_character::atk);
const param_resetter def_resetter("防御力", &cstate::def, &m100sim_character::def);
const param_resetter spd_resetter("素早さ", &cstate::spd, &m100sim_character::spd, true);

}

# define SW(name, params, spec)                                     \
  namespace                                                         \
  {                                                                 \
    EXPAND_SPEC(name, spec)                                         \
    const sinfo info_##name(sinfo_builder(effect_##name(), prio_##name).build params);\
  }                                                                 \
  const m100sim_sword_ref m100sim_s_##name = (m100sim_sword_ref)&info_##name;
# define SWN(name, kname, sname, limited, mp, spec)                 \
  namespace                                                         \
  {                                                                 \
    EXPAND_SPEC(name, spec)                                         \
    BOOST_PP_REPEAT(22, DEFINE_INFO, (name, kname, sname, limited, mp))\
  }                                                                 \
  DEFINE_INFO_FUN(name) /**/
# define SWPN(name, kname, sname, turn, limited, mp, spec)          \
  namespace                                                         \
  {                                                                 \
    EXPAND_SPEC(name, spec)                                         \
    BOOST_PP_REPEAT(22, DEFINE_INFOP, (name, kname, sname, turn, limited, mp))\
  }                                                                 \
  DEFINE_INFO_FUN(name) /**/
# define DEFINE_INFO_FUN(name)                                \
  m100sim_sword_ref m100sim_s_##name(int n)                         \
  {                                                                 \
    switch(n)                                                       \
    {                                                               \
    BOOST_PP_REPEAT(21, INFO_CASE, name)                            \
    default: return (m100sim_sword_ref)&info_##name##_21;           \
    }                                                               \
  } /**/
# define DEFINE_INFO(z, n, data) PUSHCALL(DEFINE_INFO_I, n, BOOST_PP_TUPLE_REM(5)data)
# define DEFINE_INFO_I(n, name, kname, sname, limited, mp) const sinfo info_##name##_##n(kname#n, sname#n, effect_##name(), prio_##name, n, limited, mp, boost::none);
# define PUSHCALL(f, x, im) f(x, im)
# define DEFINE_INFOP(z, n, data) PUSHCALL(DEFINE_INFOP_I, n, BOOST_PP_TUPLE_REM(6)data)
# define DEFINE_INFOP_I(n, name, kname, sname, turn, limited, mp) const sinfo info_##name##_##n(kname"."#n, sname"."#n, effect_##name(), prio_##name, turn, limited, mp, n);
# define INFO_CASE(z, n, name) case n: return (m100sim_sword_ref)&info_##name##_##n;
# define EXPAND_SPEC(name, spec) PUSHCALL(EXPAND_SPEC_I, name, BOOST_PP_TUPLE_REM(2)spec)
# define EXPAND_SPEC_I(name, prio, eff) effect effect_##name(){ eff } const int prio_##name = prio;

# define NO_SPEC (0, return effect(); )

# define ATTACK(expr) SIMPLE_ATTACK(expr, ;)
# define ATTACK_WITH(e_dmg, e_eff) SIMPLE_ATTACK(e_dmg, ADD_ATTACK_EFFECT(e_eff))
# define ATTACK_AND(e_dmg, e_add) SIMPLE_ATTACK(e_dmg, ADD_ATTACK_NEXT(e_add))
# define ATTACK_TIMES(e_dmg, k) SIMPLE_ATTACK(e_dmg, opts.count = k;)
# define SIMPLE_ATTACK(dmg, prep) (prio_atk, DEF_ATTACK_OPTS(opts, dmg) prep return single_effect(simple_attack(opts));)
# define SILENT_ATTACK(body) (prio_atk, return single_effect(rich_effect(target_enemy, simple_prim(body, 2), true));)

# define DEF_ATTACK_OPTS(name, expr) DEF_DAMAGE_FUNC(name, expr) static attack_options name(name##_damage::func);
# define DEF_DAMAGE_FUNC(name, expr) struct name##_damage { static int_t func(int turn, const cstate *self, const cstate *target) { return expr; } };
# define ADD_ATTACK_EFFECT(stmts) struct eff { static void func(gstate &gs, const effect_context &ctx, cstate *target, logger &lg) { stmts } }; opts.eff = eff::func;
# define ADD_ATTACK_NEXT(expr) struct next { static action func(gstate &gs, intparam dmg, const effect_context &ctx, cstate *target, logger &lg) { return expr; } }; opts.next = next::func;

# define PUT_BAD(desc, pos) put_bad(gs, ctx, desc, pos, target, lg);

# define AUX(body) (prio_aux, return single_effect(rich_effect(target_ally, body));)

# define CTL(msg, cond, op) (prio_ctl, DEF_CTL_FUN(msg, cond, op) return single_effect(simple_re(&ctl::func));)
# define CTL_ALL(need_sub, msg, cond, op) (prio_ctl, DEF_CTL_ALL_FUN(need_sub, msg, cond, op) return single_effect(simple_re(&ctl::func));)

# define DEF_CTL_FUN(msg, cond, op)                                 \
  struct ctl                                                        \
  {                                                                 \
    static void func(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)\
    {                                                               \
      if(!ctx.next_sword)                                           \
      {                                                             \
        lg.putf("%sには制御対象が存在しない", target->info.name.c_str());\
        return;                                                     \
      }                                                             \
      cstate *self = target;                                        \
      list<sstate> &slist = self->swords;                           \
      (void)slist; /* warning suppression */                        \
      sref it = *ctx.next_sword;                                    \
      CTL_FUN_CORE(return, msg, cond, op)                           \
    }                                                               \
  }; /**/

# define DEF_CTL_ALL_FUN(need_sub, msg, cond, op)                   \
  struct ctl                                                        \
  {                                                                 \
    static void func(gstate &gs, const effect_context &ctx, cstate *target, logger &lg)\
    {                                                               \
      if(target->swords.empty())                                    \
        lg.putf("%sには制御対象が存在しない", target->info.name.c_str());\
      if(need_sub && !ctx.next_sword) return;                       \
      sref subtarget;                                               \
      if(need_sub) subtarget = *ctx.next_sword;                     \
      cstate *self = target;                                        \
      list<sstate> &slist = self->swords;                           \
      for(sref it = slist.begin(); it != slist.end(); ++it)         \
      {                                                             \
        CTL_FUN_CORE(continue, msg, cond, op)                       \
      }                                                             \
    }                                                               \
  }; /**/

# define CTL_FUN_CORE(exit, msg, cond, op)                          \
  if(!(cond)) exit;                                                 \
  if(FILE *out = lg.start_direct())                                 \
  {                                                                 \
    fprintf(out, "%sの", target->info.name.c_str());                \
    print_sstate(out, *it);                                         \
    fputs("は", out);                                               \
    fputs((msg), out);                                              \
    lg.end_direct();                                                \
  }                                                                 \
  op; /**/

# define ADD(expr) (prio_add, DEF_CHOOSER(expr) return add_effect(chooser::func);)
# define DEF_CHOOSER(expr) struct chooser{ static boost::optional<sstate> func(gstate &gs, const effect_context &ctx, logger &lg) { return expr; } };

# define BLD(nm, hp, atk, def, spd, swords) (prio_bld, DEF_CONST_CHOOSER() SET_BUILDIA(nm, hp, atk, def, spd, swords) return build_effect(chooser::func);)
# define DEF_CONST_CHOOSER()                                        \
  struct chooser                                                    \
  {                                                                 \
    static buildia_storage &get_storage() { static buildia_storage var; return var; }\
    static const m100sim_character *func(gstate &, const effect_context &) { return &get_storage().first; }\
  }; /**/
# define SET_BUILDIA(nm, hp, atk, def, spd, swords)                 \
  character_builder st(nm, hp, atk, def, spd);                      \
  buildia_storage &var = chooser::get_storage();                    \
  st BOOST_PP_SEQ_TRANSFORM(PREFIX_SWORD, ~, swords).write(&var.first, var.second); /**/
# define PREFIX_SWORD(d, data, x) BOOST_PP_CAT(m100sim_s_, x)

# define BLD_DYN(expr) (prio_bld, DEF_BUILD_CHOOSER(expr) return build_effect(chooser::func);)
# define DEF_BUILD_CHOOSER(expr) struct chooser{ static const m100sim_character *func(gstate &gs, const effect_context &ctx) { return expr; } };

SW(slash, ("斬撃剣", "斬"),
    ATTACK(self->atk - target->def))
SW(rapid, ("高速剣", "高", 0, true),
    ATTACK(self->atk - target->def))
SW(heavy, ("重撃剣", "重", 4),
    ATTACK(2*self->atk - target->def))
SW(shock, ("衝撃剣", "衝", 2),
    ATTACK_WITH(self->atk - target->def, PUT_BAD("スタン", &status_bad::stun)))
SW(poison, ("蟲毒剣", "毒", 1, true),
    ATTACK_WITH(self->atk - target->def, PUT_BAD("毒", &status_bad::poison)))
SW(absolute, ("絶対剣", "絶", 3),
    ATTACK(self->atk))
SW(magical, ("魔法剣", "魔", 2, true),
    ATTACK(div_floor(1+target->hp,  2)))
SW(death, ("デス剣", "死", 6, true),
    ATTACK(target->hp))
SW(heat, ("加熱剣", "熱"),
    AUX(gain1("攻撃力", &cstate::atk, 1)))
SW(protect, ("加護剣", "護"),
    AUX(gain1("防御力", &cstate::def, 1)))
SW(accel, ("加速剣", "速"),
    AUX(gain1("素早さ", &cstate::spd, 1, true)))
SW(shield, ("盾の剣", "盾"),
    AUX(gain_o(guard)))
SW(mirror, ("鏡の剣", "鏡", 2, true),
    AUX(gain_o(reflection)))
SW(life, ("命の剣", "命", 1, true),
    AUX(gain_rev(life)))
SWN(recovery, "回復剣", "回", true, boost::none,
    AUX(simple_prim(&recover_full)))
SW(null, ("ヌル剣", "ヌ", 0),
    ATTACK(0))
SW(spark, ("咲火剣", "咲"),
    ATTACK(1))
SW(firecracker, ("炸撃剣", "炸"),
    ATTACK(1))
SW(eruption, ("爆炎剣", "爆", 1, false, 1),
    ATTACK(4))
SW(devour, ("貪炎剣", "貪",  1, false, 5),
    ATTACK_AND(6, recoverer(ctx.owner, dmg)))
SW(disaster, ("災厄剣", "災", 1, false, 5),
    (prio_atk,
     DEF_ATTACK_OPTS(opts0, 9)
     DEF_ATTACK_OPTS(opts1, 9)
     opts1.self = true;
     return double_effect(simple_attack(opts0), simple_attack(opts1));))
SW(perish, ("滅奏剣", "滅", 1, false, 5),
    ATTACK(1<<turn))
SW(finale, ("幕引剣", "幕", 20, true),
    ATTACK(100))
SW(combo, ("連撃剣", "連", 3),
    ATTACK_TIMES(self->atk - target->def, 2))
SW(wave, ("波状剣", "波", 6),
    ATTACK_TIMES(self->atk - target->def, 5))
SW(dance, ("舞踏剣", "舞", 3),
    (prio_atk, 
     DEF_ATTACK_OPTS(opts, self->atk - target->def)
     return double_effect(simple_attack(opts), rich_effect(target_none, simple_prim(&unuse_self, 2)));))
SW(scatter, ("全体剣", "全", 2),
    SIMPLE_ATTACK(self->atk - target->def, opts.scatter = true;))
SW(toxic, ("猛毒剣", "猛", 2, true),
    ATTACK_WITH(self->atk - target->def, PUT_BAD("猛毒", &status_bad::bad_poison)))
SW(abortion, ("堕胎剣", "堕", 2, true),
    ATTACK_WITH(self->atk - target->def, PUT_BAD("流産", &status_bad::abortion)))
SW(destroy, ("破壊剣", "壊", 1, true),
    ATTACK_WITH(self->atk - target->def, destroy_sword(gs, target, lg);))
SW(battle, ("剣戟剣", "戟", 2),
    ATTACK_WITH(self->atk - target->def, pressurize(gs, target, lg);))
SW(confusion, ("撹乱剣", "撹", 1, true),
    ATTACK_WITH(self->atk - target->def, reverse_swords(gs, target, lg);))
SW(bloodsuck, ("吸血剣", "吸", 1, true),
    ATTACK_AND(self->atk - target->def, recoverer(ctx.owner, dmg)))
SW(robbery, ("強盗剣", "盗", 2, true),
    ATTACK_AND(self->atk - target->def, dmg > 0 ? sword_stealer(ctx.owner, target) : action(no_action)))
SW(dark, ("闇黒剣", "闇", 1, false, 2),
    SIMPLE_ATTACK(self->atk - target->def, opts.penetrate = true;))
SW(antimagic, ("魔封剣", "封"),
    SIMPLE_ATTACK(self->atk - target->def, opts.mp = true;))
SW(overweight, ("過重剣", "過", 5),
    ATTACK(3*self->atk - target->def))
SW(solid, ("確実剣", "確"),
    ATTACK(self->atk / 2))
SW(sonic, ("音速剣", "音", 0, true),
    ATTACK_TIMES(self->atk / 5 - target->def, 10))
SW(light, ("光速剣", "光", 0, true),
    ATTACK_TIMES(self->atk / 10 - target->def, 100))
SWN(rage, "激怒剣", "怒", true, boost::none,
    ATTACK(self->max_hp - self->hp))
SW(cool, ("冷却剣", "冷", 1, false, 1),
    SILENT_ATTACK(atk_resetter))
SW(decay, ("剥落剣", "剥", 1, false, 1),
    SILENT_ATTACK(def_resetter))
SW(decel, ("減速剣", "減", 1, false, 1),
    SILENT_ATTACK(spd_resetter))
SW(gentle, ("優撃剣", "優", 2, true),
    SILENT_ATTACK(&recover_full))
SW(flashfire, ("炎上剣", "炎", 1, false, 5),
    (prio_atk,
     return double_effect(
       rich_effect(target_enemy, simple_prim(&do_flashfire, 2), true),
       rich_effect(target_none,  simple_prim(&do_flashfire, 2), false));))
SW(alert, ("警戒剣", "警", 0, true),
    AUX(gain_o(guard)))
SW(evasion, ("回避剣", "避", 1, true),
    AUX(gain_o(detection)))
SW(escutcheon, ("楯の剣", "楯",  1, false, 1),
    AUX(gain_o(shield)))
SW(amnion, ("羊膜剣", "膜",  1, false, 2),
    AUX(gain_o(cusion)))
SW(barrier, ("障壁剣", "壁",  1, false, 3),
    AUX(gain_o(barrier)))
SW(soul, ("魂の剣", "魂", 3, true),
    AUX(gain_rev(soul)))
SW(push, ("押分剣", "押", 0),
    AUX(gain_os("プッシュ", &status_other::push)))
SW(drive, ("駆動剣", "駆"),
    AUX(gain_os("ドライブ", &status_other::drive)))
SW(wriggle, ("蠕動剣", "蠕"),
    AUX(gain_o(flip)))
SW(attraction, ("誘引剣", "誘", 1, true),
    AUX(gain_os("挑発", &status_other::provocation)))
SW(first, ("先行剣", "先", 0, true),
    AUX(gain_os("クイック", &status_other::quick, true)))
SW(second, ("後続剣", "後", 0, true),
    AUX(gain_os("スロー", &status_other::slow, true)))
SW(communication, ("疎通剣", "疎", 0, true),
    AUX(gain_os("ヘヴンズドア", &status_other::heavens_door)))
SWN(heal, "快癒剣", "癒", true, boost::none,
    AUX(simple_prim(&recover_all)))
SW(chant, ("詠唱剣", "詠"),
    AUX(gain1("MP", &cstate::mp, 1)))
SW(prayer, ("黙祷剣", "黙", 1, true),
    AUX(gain1("MP", &cstate::mp, 3)))
SWN(ritual, "儀式剣", "儀", true, boost::none,
    AUX(gain1("AP", &cstate::ap, 3, true)))
SW(prospect, ("臥竜剣", "臥", 0, true),
    AUX(simple_prim(&do_prospect)))
SW(dragonize, ("竜化剣", "竜", 10, true),
    AUX(simple_prim(&do_dragonize)))
SW(limit, ("限定剣", "限", 0, true),
    CTL("リミテッドを得た", true, it->ench.limited = true))
SW(suppression, ("抑制剣", "抑", 0),
    CTL("使用済みになった", 2 * self->hp >= self->max_hp, it->used = true))
SWPN(delay, "遅延剣", "遅", 0, false, boost::none,
    CTL("使用済みになった", true, it->used = true))
SWN(sheathe, "収鞘剣", "鞘", true, boost::none,
    CTL("消滅した", true, slist.erase(it)))
SWN(derive, "導来剣", "導", true, boost::none,
    CTL("早期起動を得た", true, it->ench.early++))
SW(vibration, ("振動剣", "振", 1, true),
    CTL("スタン与えを得た", it->info->prio == prio_atk, it->ench.stun++))
SW(toxicant, ("毒薬剣", "薬", 1, true),
    CTL("毒与えを得た", it->info->prio == prio_atk, it->ench.poison++))
SW(diffuse, ("拡散剣", "散", 1, true),
    CTL("全体化を得た", it->info->prio == prio_atk || it->info->prio == prio_aux, it->ench.scatter = true))
SW(train, ("鍛錬剣", "鍛", 4, true),
    CTL_ALL(false, "重撃剣になった", it->info == sinfo_by_ref(m100sim_s_slash), change_sword_type(&*it, sinfo_by_ref(m100sim_s_heavy))))
SWN(recurrence, "再来剣", "再", true, boost::none,
    CTL_ALL(false, "未使用になった", true, it->used = false))
SWN(resonance, "共鳴剣", "鳴", true, boost::none,
    CTL_ALL(true,
      (print_sstate(out, *subtarget), "になった"),
      true,
      change_sword_type_as(&*it, *subtarget)))
SW(fictional, ("虚構剣", "虚"),
    ADD(sstate(m100sim_s_null)))
SW(multi, ("増殖剣", "増", 1, false, 1),
    ADD(ctx.next_sword ? **ctx.next_sword : boost::optional<sstate>()))
SW(replica, ("複製剣", "複", 1, true),
    ADD(gs.teams[opponent(ctx.owner->info.side)].chs.front().swords.empty()
      ? boost::optional<sstate>()
      : gs.teams[opponent(ctx.owner->info.side)].chs.front().swords.front()))
SW(reverb, ("残響剣", "残", 0, true, 5),
    BLD_DYN(gs.teams[ctx.owner->info.side].primary))
SW(echo, ("反響剣", "反", 0, true, 5),
    BLD_DYN(gs.teams[opponent(ctx.owner->info.side)].primary))
SW(bubble, ("泡溢剣", "泡", 0, false, 1),
    BLD("バブリア", 5, 0, 0, 0, (none)))
SW(spirit, ("精霊剣", "精", 0, false, 4),
    BLD("ウィスプ", 5, 0, 0, 1, (spirit)(chant)))
SW(fetal, ("胎胎剣", "胎", 0, true, 3),
    BLD("マトリョックス", 15, 0, 0, 1, (prayer)(fetal)))
SW(sawhair, ("鋸髪剣", "鋸", 0, true, 5),
    BLD("デスエリーナ", 5, 0, 0, 3, (accel)(protect)(protect)(shock)(shock)(death)))
SW(sawa, ("相沢剣", "沢", 0, true, 5),
    BLD("あいざわ", 15, 0, 4, 1, (slash)(death)))
SW(purple, ("紫電剣", "紫", 0, true, 5),
    BLD("紫電", 5, 0, 0, 4, (accel)(shock)(heat)(shock)(heat)))
SW(holand, ("尾蘭剣", "蘭", 0, true, 5),
    BLD("無考オランダ", 5, 3, 0, 3, (death)(death)(death)))
SW(servant, ("従僕剣", "従", 0, true, 5),
    BLD("しゃるる", 5, 3, 0, 2, (rapid)(rapid)(poison)(poison)))
SW(leg, ("弩脚剣", "脚", 0, true, 5),
    BLD("るーしえ君", 40, 0, 0, 1, (death)))
SW(dog, ("狗咬剣", "狗", 0, true, 5),
    BLD("徘徊するライカ", 15, 0, 1, 2, (accel)(heat)(shock)(absolute)))
SW(ko, ("ｺｳ速剣", "ｺｳ", 0, true, 5),
    BLD("コウ", 5, 5, 0, 2, (rapid)(rapid)))
SW(yoshi, ("吉宏剣", "吉", 0, true, 5),
    BLD("武本吉宏", 10, 0, 0, 3, (accel)(heat)(heat)(heat)(heavy)))
SW(holy, ("聖重剣", "聖", 0, true, 6),
    BLD("ゾアラ", 15, 0, 0, 3, (accel)(protect)(heat)(heat)(heavy)))
SW(geki, ("撃鉄剣", "撃", 0, true, 5),
    BLD("撃鉄", 10, 0, 0, 4, (chant)(eruption)(chant)(eruption)))
SW(emperor, ("鶏皇剣", "鶏", 0, true, 13),
    BLD("撃鉄Act.6", 55, 0, 3, 6, (prayer)(prayer)(suppression)(perish)(delay(19))(perish)))
SW(bone, ("怪骨剣", "怪", 0, true, 7),
    BLD("メテオラ", 10, 5, 3, 2, (slash)(slash)))
SW(princess, ("歪姫剣", "歪", 0, true, 5),
    BLD("司闇", 10, 0, 0, 2, (prayer)(prayer)(reverb)(death)(chant)(amnion)))
SW(summon, ("招龍剣", "龍", 0, true, 10),
    BLD("ドラゴン", 30, 0, 6, 4, (heat)(slash)(heat)(slash)))
SW(age, ("星霜剣", "星", 1, false, 1),
    (prio_env, return single_effect(simple_re(&do_age));))

SW(recoveryA, ("回復剣", "回", 1, true, boost::none, boost::none, true),
    AUX(simple_prim(&recover_full)))
SW(perishA, ("滅奏剣", "滅",  1, true, boost::none, boost::none, true),
    ATTACK(1<<turn))
SW(rageA, ("激怒剣", "怒", 1, true, boost::none, boost::none, true),
    ATTACK(self->max_hp - self->hp))
SW(fang, ("牙撃剣", "牙", 2),
    ATTACK(2*self->atk - target->def))
SW(two, ("二撃剣", "二", 2),
    ATTACK_TIMES(self->atk - target->def, 2))
SW(three, ("三撃剣", "三", 3),
    ATTACK_TIMES(self->atk - target->def, 3))
SW(four, ("四撃剣", "四", 4),
    ATTACK_TIMES(self->atk - target->def, 4))
SW(five, ("五撃剣", "五", 5),
    ATTACK_TIMES(self->atk - target->def, 5))
SW(six, ("六撃剣", "六", 6),
    ATTACK_TIMES(self->atk - target->def, 6))
SW(seven, ("七撃剣", "七", 7),
    ATTACK_TIMES(self->atk - target->def, 7))
SW(abortionA, ("堕胎剣", "堕", 1, true),
    ATTACK_WITH(self->atk - target->def, PUT_BAD("流産", &status_bad::abortion)))
SW(coolA, ("冷却剣", "冷", 2),
    ATTACK_WITH(self->atk - target->def, atk_resetter(gs, ctx, target, lg);))
SW(decayA, ("剥落剣", "剥", 2),
    ATTACK_WITH(self->atk - target->def, def_resetter(gs, ctx, target, lg);))
SW(decelA, ("減速剣", "減", 2),
    ATTACK_WITH(self->atk - target->def, spd_resetter(gs, ctx, target, lg);))
SW(darkA, ("闇黒剣", "闇",  2),
    SIMPLE_ATTACK(self->atk - target->def, opts.penetrate = true;))
SW(antimagicA, ("魔封剣", "封", 1, true),
    SIMPLE_ATTACK(self->atk - target->def, opts.mp = true;))
SW(pushA, ("押分剣", "押", 0, true),
    AUX(gain_os("プッシュ", &status_other::push)))
SW(attractionA, ("誘引剣", "誘", 0, true),
    AUX(gain_os("挑発", &status_other::provocation)))
SW(healA, ("快癒剣", "癒", 1, true, boost::none, boost::none, false, true),
    AUX(simple_prim(&recover_all)))
SW(chantA, ("詠唱剣", "詠", 0),
    AUX(gain1("MP", &cstate::mp, 1)))
SW(rogation, ("祈祷剣", "祈", 0, true),
    AUX(gain1("MP", &cstate::mp, 2)))
SW(ritualA, ("儀式剣", "儀", 0, true),
    AUX(gain1("予備AP", &cstate::reserved_ap, 3, true)))
SW(blaze, ("灼熱剣", "灼", 3, true),
    AUX(simple_prim(&double_attack)))
SW(deriveA, ("導来剣", "導", 0, true),
    CTL("早期起動を得た", true, it->ench.early++))
SW(slow, ("遅滞剣", "遅", 0, true),
    CTL("遅滞起動を得た", true, it->ench.early--))
SW(danger, ("危殆剣", "危", 0, true),
    CTL("危殆起動を得た", true, it->ench.reacts_to_danger = true))
SW(recurrenceA, ("再来剣", "再", 3, true),
    CTL_ALL(false, "未使用になった", true, it->used = false))
SW(resonanceA, ("共鳴剣", "鳴", 4, true),
    CTL_ALL(true,
      (print_sstate(out, *subtarget), "になった"),
      true,
      change_sword_type_as(&*it, *subtarget)))
SW(ageA, ("星霜剣", "星"),
    (prio_env, return single_effect(simple_re(&do_age));))
SW(mkbubble, ("造泡剣", "泡", 0, true, 2),
    BLD("バブル", 5, 0, 0, 0, (none)))
SW(mkfence, ("造柵剣", "柵", 0, true, 3),
    BLD("フェンス", 20, 0, 0, 0, (none)))
SW(mkgate, ("造門剣", "門", 0, true, 4),
    BLD("ゲート", 35, 0, 0, 0, (none)))
SW(mkwall, ("造壁剣", "壁", 0, true, 5),
    BLD("ウォール", 50, 0, 0, 0, (none)))
SW(mkice, ("造氷剣", "氷", 0, true, 3),
    BLD("アイシア", 10, 0, 0, 1, (mirror)))
SW(mkmist, ("造霧剣", "霧", 0, true, 4),
    BLD("ミスティア", 15, 0, 0, 2, (diffuse)(mirror)))
SW(mktear, ("造涙剣", "涙", 0, true, 5),
    BLD("ウンディーネ", 20, 0, 0, 3, (soul)(diffuse)(mirror)))

SWN(victory, "勝利剣", "勝", true, boost::none,
    AUX(gain_os("重プッシュ", &status_other::heavy_push, false)))
