{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   'IDevice' and 'ODevice' represent any files other than regular
   files, including standard input\/output.
-}

module System.IO.SSC.Device
  ( IDevice
  , ODevice
  , openIDevice
  , openODevice
  , iDeviceFromFd
  , oDeviceFromFd
  , DeviceInfo
  , sWithFd
  ) where

import Control.Monad
import Data.Word
import System.IO(IOMode(..))
import System.IO.SSC.Classes
import System.IO.SSC.Internal.BasicPort
import System.IO.SSC.Internal.Fd
import System.IO.SSC.Internal.Debug

newtype IDevice = ID (BasicPort Fd)
newtype ODevice = OD (BasicPort Fd)
newtype DeviceInfo = DI Fd

instance Port IDevice where
  rawClose (ID bp) = bpClose bp fdClose
  rawIsClosed (ID bp) = bpIsClosed bp
  rawFlush _ = return ()

instance Port ODevice where
  rawClose (OD bp) = bpClose bp fdClose
  rawIsClosed (OD bp) = bpIsClosed bp
  rawFlush _ = return ()

instance Destination DeviceInfo IDevice where
  rawDest (ID bp) = return $ DI $ bpBase bp

instance Destination DeviceInfo ODevice where
  rawDest (OD bp) = return $ DI $ bpBase bp

instance Lockable IDevice where
  sLock (ID bp) = bpLock bp
  sTryLock (ID bp) = bpTryLock bp
  sUnsafeUnlock (ID bp) = bpUnsafeUnlock bp
  
instance Lockable ODevice where
  sLock (OD bp) = bpLock bp
  sTryLock (OD bp) = bpTryLock bp
  sUnsafeUnlock (OD bp) = bpUnsafeUnlock bp

instance IBlock Word8 IDevice where
  rawGetBufB (ID bp) bmode buf size = do
    msg $ "+rawGetBufB/IDevice size=" ++ show size ++ "; mode=" ++ show bmode
    bpEnsureOpen bp
    r <- fdGetBufB (bpBase bp) bmode buf size
    msg $ "-rawGetBufB/IDevice r=" ++ show r
    return r

instance OBlock Word8 ODevice where
  rawPutBufB (OD bp) bmode buf size = do
    bpEnsureOpen bp
    fdPutBufB (bpBase bp) bmode buf size

-- | Open an input device specified by a string, whose meaning is
-- platform-dependent.
openIDevice :: String -> IO IDevice
openIDevice name = do
  fd <- openRawBinaryFd name ReadMode
  liftM ID $ basicPort fd 

-- | Open an output device specified by a string, whose meaning is
-- platform-dependent.
openODevice :: String -> IO ODevice
openODevice name = do
  fd <- openRawBinaryFd name WriteMode
  liftM OD $ basicPort fd 

iDeviceFromFd :: Fd -> IO IDevice
iDeviceFromFd fd = do
  fdSetFlags fd
  liftM ID $ basicPort fd

oDeviceFromFd :: Fd -> IO ODevice
oDeviceFromFd fd = do
  fdSetFlags fd
  liftM OD $ basicPort fd

sWithFd :: (Destination DeviceInfo s) => s -> (Fd -> IO a) -> IO (Maybe a)
sWithFd s f = do
  DI fd <- rawDest s
  liftM Just $ f fd
